Sackville = LibStub("AceAddon-3.0"):NewAddon("Sackville", "AceConsole-3.0", "AceHook-3.0", 'AceEvent-3.0')

local sprintf = string.format
local _
local debugVerbosity = 0
local debugFlags = {
    init = false,
    baglayout = false,
    status = false,
    scan = false,
    item = true,
    filter = false,
    event = false,
    layout = false
}

-- TODO: collect all options/defaults somewhere
local COLUMN_WIDTH = 40
local ROW_HEIGHT = 40
local BUTTON_HEIGHT = 40
local BUTTON_WIDTH = 40
local FRAME_PADDING = 0

local loading = true
local bankInitialized = false
local isBankOpen = false

local db = nil
local bagIndexes = {}
local dummyFrame = nil

Sackville.items = {}
Sackville.playerBags = {}
Sackville.bankBags = {}
Sackville.quickPanel = nil
Sackville.emptyBagSlot = {
    bank = nil,
    player = nil,
}

function Sackville:DumpItem(item)
    if item then
        local id = item.id or "?"
        local name = item.name or "?"
        local bagslot = item.bagslot or "?"
        local itemType = item.type or "?"
        local itemSubType = item.subType or "?"
        local removed = item.removed and "true" or "false"
        return sprintf("id: %s, name: %s, bagslot: %s, type: %s, subType: %s, itemRemoved: %s",
            id, name, bagslot, itemType, itemSubType, removed)
    else
        return "nil item"
    end

end

local function PositionFrame(bagFrame, state)
    -- local bagFrame = vbag:GetFrame()
    local h = bagFrame:GetHeight()
    local w = bagFrame:GetWidth()
    Sackville:Debug(2, "baglayout",
        " PositionFrame, x_offset: %d, y_offset: %d, widest_bag: %d, w: %d, h: %d, SH: %d",
        state.x_offset, state.y_offset, state.widest_bag, w, h, GetScreenHeight())
    state.widest_bag = (w > state.widest_bag) and w or state.widest_bag
    
    if (state.y_offset + h) < GetScreenHeight() then
        -- There is enough room to stack this frame vertically
        Sackville:Debug(2, "baglayout", "  Setting pos to (%d, %d)",
            state.x_offset, state.y_offset)
        bagFrame:SetPoint("BOTTOMRIGHT", state.x_offset, state.y_offset)
        state.y_offset = state.y_offset + h + FRAME_PADDING
    else
        -- Start a new column
        state.y_offset = 80
        state.x_offset = state.x_offset - state.widest_bag - FRAME_PADDING
        state.widest_bag = 0
        Sackville:Debug(2, "baglayout", "  Setting pos to (%d, %d)", 
            state.x_offset, state.y_offset)
        bagFrame:SetPoint("BOTTOMRIGHT", state.x_offset, state.y_offset)
        state.y_offset = state.y_offset + h + FRAME_PADDING
    end
    
    return state
end

function Sackville:IsBankBagId(bagId)
    for _, bankBagId in ipairs(bagIndexes.bank) do
        if bagId == bankBagId then
            return true
        end
    end
    
    return false
end

function Sackville:IsPlayerBagId(bagId)
    for _, playerBagId in ipairs(bagIndexes.bags) do
        if bagId == playerBagId then
            return true
        end
    end
    
    return false
end

function BuildBagIndexes()
   bagIndexes = {
      bags = {},
      bank = {}
   }

   for i = 0, NUM_BAG_SLOTS do
      table.insert(bagIndexes.bags, i)
   end
   
   table.insert(bagIndexes.bank, BANK_CONTAINER)
   for i = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
      table.insert(bagIndexes.bank, i)
   end
end

function Sackville:GetMoneyString()
   local copper = GetMoney()
   local money

   local size = 12
   local t = sprintf('%d|TInterface\\MoneyFrame\\UI-GoldIcon:%d:%d:3:0:64:64:4:60:4:60|t ', copper / 100 / 100, size, size)
   t = t..sprintf(' %d|TInterface\\MoneyFrame\\UI-SilverIcon:%d:%d:3:0:64:64:4:60:4:60|t ', (copper / 100) % 100, size, size)
   t = t..sprintf(' %d|TInterface\\MoneyFrame\\UI-CopperIcon:%d:%d:3:0:64:64:4:60:4:60|t ', copper % 100, size, size)

   for i=1, GetCurrencyListSize() do
      local name, isHeader, _, _, isWatched, count, icon = GetCurrencyListInfo(i)
      if (not isHeader) then
         if isWatched then
            t = (count or 0)..'|T'..icon..':'..size..':'..size..':3:0:64:64:4:60:4:60|t   '..t
         end
         if money then
            money[i] = count
         end
      end
   end

   return t
end

function Sackville:RebuildInventory(updatedBagId)
-- Scans a specific bag (or all bags if none specified) and 
-- populates/updates self.items, returning a table of all items that
-- have been added/changed/deleted.
    local bagid = updatedBagId or "nil"
    self:Debug(1, "status", "RebuildInventory: %s", bagid)
    local result = nil    
    local bagType = nil
    
    if updatedBagId ~= nil then
        if self:IsPlayerBagId(updatedBagId) then
            bagType = "player"
        elseif self:IsBankBagId(updatedBagId) then
            bagType = "bank"
        else
            assert(bagType, "Unknown bag type for bagid: %s ", bagid)
            -- Support future bag types (guild vault, void storage)
        end

        if bagType ~= nil then -- Scan a specific bag
            local slots = GetContainerNumSlots(updatedBagId)
            if slots then
                result = Sackville:ScanBagSlots(updatedBagId, slots, result)
            else
                self:Debug(2, "scan", "Bag %s has no slots", bagid)
            end
        end
    else -- Scan everything
        for _, bagid in ipairs(bagIndexes.bags) do
            local slots = GetContainerNumSlots(bagid)
            if slots then
                result = Sackville:ScanBagSlots(bagid, slots, result)
            else
                self:Debug(2, "scan", "Player bag %s has no slots", bagid)
            end
        end
        
        if isBankOpen then
            for _, bagid in ipairs(bagIndexes.bank) do
                local slots = GetContainerNumSlots(bagid)
                if slots then
                    result = Sackville:ScanBagSlots(bagid, slots, result)
                else
                    self:Debug(2, "scan", "Bank bag %s has no slots", bagid)
                end
            end
        end
    end
    
    return result
end

-- TODO: Add a description
function Sackville:ScanBagSlots(bagid, slots, priorResult)
    self:Debug(2, "scan", "  ScanBagSlots (" .. bagid .. ", " .. slots .. ")")

    local result = {}
    
    if priorResult then result = priorResult
    else
        result.itemsAdded = {}
        result.itemsChanged = {}
        result.itemsRemoved = {}    
    end
    
    for slot = 1, slots do
        local bagslot = bagid .. '-' ..slot
        local itemid = GetContainerItemID(bagid, slot)
        if itemid then -- slot is not empty
            local isNewItem, itemHasChanged = self:LoadItem(bagid, slot, itemid)
            
            if isNewItem then 
                table.insert(result.itemsAdded, self.items[bagslot])           
            elseif itemHasChanged then
                table.insert(result.itemsChanged, self.items[bagslot])
            end
        else
            if self.items[bagslot] and not self.items[bagslot].removed then
                table.insert(result.itemsRemoved, self.items[bagslot])
            end
        end
    end
    
    self:Debug(2, "status", "  # of items added: %d, changed: %d, removed: %d",
        #result.itemsAdded, #result.itemsChanged, #result.itemsRemoved)
    return result
end

function Sackville:LoadItem(bagid, slot, itemid)
-- Builds an item and assigns it to self.items[bagslot], 
-- or updates an existing item
    assert(itemid, "LoadItem: nil id at bagslot " .. bagid .. '-' .. slot)

    local isNewItem = nil
    local itemHasChanged = nil    
    local itemCount 
    _, itemCount = GetContainerItemInfo(bagid, slot)

    local savedItem = self.items[bagid .. "-" .. slot]
    self:Debug(3, "item", "LoadItem(%s, %s, %s, savedItem=[%s])",
        bagid, slot, itemid, self:DumpItem(savedItem))
        
    if savedItem == nil then
        isNewItem = true
        itemHasChanged = false
    else
        if savedItem.removed then
            isNewItem = true
            itemHasChanged = true
        elseif savedItem.id == itemid and savedItem.count == itemCount then
            isNewItem = false
            itemHasChanged = false
        else
            isNewItem = false
            itemHasChanged = true
        end
    end
    
    local itemName, _, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType,
        itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemid) 
    local isQuestItem, questID, isActive = GetContainerItemQuestInfo(bagid, slot)
    
    local currentItem = savedItem or {}
    
    if isNewItem or itemHasChanged then
        currentItem.id = itemid
        currentItem.count = itemCount
        currentItem.name = itemName or "Unknown item"
        currentItem.quality = itemRarity
        currentItem.level = itemLevel
        currentItem.minLevel = itemMinLevel
        currentItem.equipLoc = itemEquipLoc
        currentItem.type = itemType
        currentItem.subType = itemSubType
        currentItem.stackCount = itemStackCount
        currentItem.texture = itemTexture
        currentItem.bagid = bagid
        currentItem.slot = slot
        currentItem.bagslot = bagid .. "-" .. slot  
        currentItem.isQuestItem = isQuestItem
        currentItem.questID = questID
        currentItem.isActive = isActive
            
        if isNewItem then
            self:Debug(3, "item", "   Creating new button for %s", 
                self:DumpItem(currentItem))
            currentItem.button = self:createButtonFromItem(currentItem)
        else
            self:UpdateItemButton(currentItem)
        end

        currentItem.removed = false
        self.items[currentItem.bagslot] = currentItem
    end
    
    self:Debug(3, "item", "Loaded item: %s", self:DumpItem(currentItem))
    return isNewItem, itemHasChanged
end

function Sackville:createButtonFromItem(item)
    -- This wrapper frame allows the calls to self:GetParent():GetID() to
    -- function properly in the inherited Blizzard UI code, since the
    -- parent ID is expected to be the bagid and the child ID is the slot.
    local basename
    local buttonTemplate 
    
    if item.bagid == -1 then
        basename = "SackvilleBankItemButton" .. item.bagslot
        buttonTemplate = "BankItemButtonGenericTemplate"
    else
        basename = "SackvilleItemButton" .. item.bagslot
        buttonTemplate = "ContainerFrameItemButtonTemplate"
    end
    
    wrapper = _G[basename.."_wrapper"]
        or CreateFrame("Frame", basename.."_wrapper", dummyFrame)
    wrapper:ClearAllPoints()
    wrapper:SetPoint("TOPLEFT")
    wrapper:SetID(item.bagid)
    wrapper:SetHeight(BUTTON_HEIGHT)
    wrapper:SetWidth(BUTTON_WIDTH)
    wrapper:Show()
    -- 
    item.button = _G[basename] or 
        CreateFrame("Button", basename, wrapper, buttonTemplate)
    item.button:SetAllPoints() -- fill the parent frame
    item.button:SetID(item.slot)
    SetItemButtonTexture(item.button, item.texture)
    SetItemButtonCount(item.button, item.count)
    self:SetQualityOverlay(item)
    self:UpdateButtonCooldown(item)
    item.button:Show()
    item.button.wrapper = wrapper

    self:Debug(3, "item", "Created button %s, SetID(%d)", 
        item.button:GetName(), item.slot)
    return item.button
end

function Sackville:SetQualityOverlay(item)
-- This will overlay a frame with a colored border over the button.
    local basename = "SackvilleItemButton" .. item.bagslot .. "_quality"
    local button = item.button
    button.qualityFrame = _G[basename]
    or CreateFrame("Frame", basename, item.button)

    
    if (item.quality == nil) or (item.quality < 2) or (item.quality > 7) then 
        button.qualityFrame:Hide()
    else
        local questTexture = (item.questID and not item.isActive) 
        and TEXTURE_ITEM_QUEST_BANG or (item.questID or item.isQuestItem) 
        and TEXTURE_ITEM_QUEST_BORDER
        
        button.qualityFrame.texture = _G[basename .. '_texture']
            or button.qualityFrame:CreateTexture(basename .. '_texture', 'OVERLAY')
            
        if questTexture then
            button.qualityFrame.texture:SetTexture(questTexture)
        else
            button.qualityFrame.texture:SetTexture(sprintf(
            'Interface\\Addons\\Sackville\\Textures\\UI-Icon-Quality%sBorder.blp',
            item.quality))
        end
        button.qualityFrame.texture:SetBlendMode('BLEND')
        button.qualityFrame.texture:SetAllPoints()
        button.qualityFrame.texture:Show()
        button.qualityFrame:SetAllPoints()
        button.qualityFrame:Show()
    end
end

function Sackville:UpdateButtonCooldown(item, index)
    if item then
        index = index or ""
        assert(item.button, "UpdateButtonCooldown " .. index .. 
            ": nil button: " .. self:DumpItem(item))
        
        local cooldown = _G[item.button:GetName().."Cooldown"]
        local start, duration, enable = GetContainerItemCooldown(item.bagid, item.slot)
        CooldownFrame_SetTimer(cooldown, start, duration, enable);
        if ( duration > 0 and enable == 0 ) then
            SetItemButtonTextureVertexColor(item.button, 0.4, 0.4, 0.4);
        else
            SetItemButtonTextureVertexColor(item.button, 1, 1, 1);
        end
    end
end

function Sackville:UpdateItemButton(item)
    self:Debug(3, "item", "Updating button for %s with id %d at %s",
        item.name, item.id, item.bagslot)

    SetItemButtonTexture(item.button, item.texture)
    SetItemButtonCount(item.button, item.count)
    self:SetQualityOverlay(item)
end

function Sackville:OnInitialize()
-- Code that you want to run when the addon is first loaded goes here.
-- do init tasks here, like loading the Saved Variables, 
-- or setting up slash commands.

    -- self.db = LibStub("AceDB-3.0"):New("SackvilleDB", defaults, 'Default')
    -- db = self.db.profile
    Sackville:RegisterChatCommand("sack", "ChatCommand")
    
    self.quickPanel = Sackville.VirtualBag:CreateQuickPanel()
end

function Sackville:OnEnable()
-- Do more initialization here, that really enables the use of your addon.
-- Register Events, Hook functions, Create Frames, Get information from 
-- the game that wasn't available in OnInitialize
    -- self:RegisterEvent('BAG_UPDATE')
    self:RegisterEvent('PLAYER_ENTERING_WORLD')
    -- self:RegisterEvent("UNIT_INVENTORY_CHANGED")
    -- self:RegisterEvent('ITEM_LOCK_CHANGED')
    -- self:RegisterEvent('BAG_UPDATE_COOLDOWN')
    -- self:RegisterEvent('EQUIPMENT_SETS_CHANGED')
    self:RegisterEvent("BANKFRAME_OPENED")
    self:RegisterEvent("BANKFRAME_CLOSED")
    -- self:RegisterEvent("PLAYER_MONEY")
    -- self:RegisterEvent('AUCTION_HOUSE_SHOW')
    -- self:RegisterEvent('AUCTION_HOUSE_CLOSED')
    -- self:RegisterEvent("CURRENCY_DISPLAY_UPDATE");
    -- self:RegisterEvent("KNOWN_CURRENCY_TYPES_UPDATE");
    self:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
   self:RawHook("ToggleAllBags", "ToggleBags", true)
   self:RawHook("OpenAllBags", "ShowBags", true)
   self:RawHook("OpenBackpack", "ShowBags", true)
   --self:RawHook("CloseBackpack", "HideBags", true)
   --self:RawHook("CloseSpecialWindows", "HideBagsOnEsc", true)    
   -- self:RawHook("SetCurrencyBackpack", "SetCurrencyBackpack", true)
   -- self:RawHookScript(BankFrame, "OnEvent", "BankFrameOnEvent")
   
   -- self:RegisterEvent("CURRENCY_DISPLAY_UPDATE")
   self:RegisterEvent("PLAYER_MONEY", "UpdateMoney")
   -- self:RegisterEvent("PLAYER_TRADE_MONEY", "UpdateMoney")
   -- self:RegisterEvent("TRADE_MONEY_CHANGED", "UpdateMoney")
   -- self:RegisterEvent("SEND_MAIL_MONEY_CHANGED", "UpdateMoney")
   -- self:RegisterEvent("SEND_MAIL_COD_CHANGED", "UpdateMoney")
   -- self:RegisterEvent("GUILDBANK_UPDATE_MONEY", "UpdateMoney")
   -- self:RegisterEvent("GUILDBANK_UPDATE_WITHDRAWMONEY", "UpdateMoney")
   self:RegisterEvent("CURRENCY_DISPLAY_UPDATE", "UpdateMoney")
   -- self:RegisterEvent("", "UpdateMoney")
   -- self:RegisterEvent("", "UpdateMoney")
   -- self:RegisterEvent("", "UpdateMoney")
   
    dummyFrame = CreateFrame("Frame", "SackvilleDummyFrame", UIParent)   
    self:InitializeBags()
end


function Sackville:OnDisable()
-- Unhook, Unregister Events, Hide frames that you created.
-- You would probably only use an OnDisable if you want to 
-- build a "standby" mode, or be able to toggle modules on/off.
    self:UnregisterEvent('PLAYER_ENTERING_WORLD')
    self:UnregisterEvent("BANKFRAME_OPENED")
    self:UnregisterEvent("BANKFRAME_CLOSED")
    self:UnregisterEvent("PLAYERBANKSLOTS_CHANGED")

    self:RawHook("ToggleAllBags", "ToggleBags", true)
    self:RawHook("OpenAllBags", "ShowBags", true)
    self:RawHook("OpenBackpack", "ShowBags", true)

    self:HideBags()
end

function Sackville:InitializeBags()
    -- TODO: First clear all existing sections in case this is called a more than once
    for bagIndex=1, #self.Options.bagConfig do
        local bagOptions = self.Options.bagConfig[bagIndex]
        local vbag = Sackville.VirtualBag:New(bagOptions, bagIndex)

        -- for sectionName, sectionOptions in pairs (bagOptions.sections) do        
        for sectionIndex=1, #bagOptions.sections do
            vbag:AddSection(bagOptions.sections[sectionIndex], sectionIndex)
        end
        
        if vbag.type == "player" then
            self.playerBags[bagOptions.name] = vbag
        elseif vbag.type == "bank" then
            self.bankBags[bagOptions.name] = vbag
        else
            -- Support some other bag type
        end
        assert(vbag.name)
        self:Debug(1, "init", "Initialized vbag (%s)", vbag.name)
    end
end

function Sackville:PLAYER_ENTERING_WORLD()
    BuildBagIndexes()
    local result = self:RebuildInventory()

    for _, item in ipairs(result.itemsAdded) do
        local section = self:GetVbagSectionForItem(item, self.playerBags, "player")
        if section then
            section.vbag:AddItem(item, section.name)
        else
            self:Debug(3, "filter", "Couldn't find a bag for %s", item.name)
        end
    end

    loading = false
end

function Sackville:BAG_UPDATE(event, bagid)
    if (loading) then return end    
    self:Debug(1, "event", "BAG_UPDATE: %s", bagid)

    local bagSet
    local bagType

    if self:IsPlayerBagId(bagid) then
        bagSet = self.playerBags
        bagType = "player"

    elseif (isBankOpen and self:IsBankBagId(bagid)) then
        bagSet = self.bankBags
        bagType = "bank"
    else
        assert(bagSet, "Unknown bagtype for bagid: " .. bagid)
        -- Blew up with bagid 5 after porting
    end

    local result
	if bagSet then
        result = self:RebuildInventory(bagid)
    
        for _, item in ipairs(result.itemsAdded) do
            local section = self:GetVbagSectionForItem(item, bagSet, bagType)
            if section then
                section.vbag:AddItem(item, section.name)
            else
                self:Debug(1, "filter", "Couldn't find a bag for %s", item.name)
            end            
        end
        
        for _, item in ipairs(result.itemsRemoved) do
            item.removed = true
            item.section.vbag:RemoveItem(item)
            item.button.wrapper:SetParent(dummyFrame)
        end
        
        for _, item in ipairs(result.itemsChanged) do
            item.button.wrapper:Show()
            local section = self:GetVbagSectionForItem(item, bagSet, bagType)
            if section then                
                if (section.vbag.name) ~= item.section.vbag.name 
                or (section.name ~= item.section.name) then
                    item.section.vbag:RemoveItem(item)
                    section.vbag:AddItem(item, section.name)
                end
            else
                self:Debug(1, "filter", "Couldn't find a bag for %s", item.name)
            end            

        end
    end

    -- No need to refresh the layout immediately if we only removed items.
    -- if result and #result.itemsAdded > 0 or #result.itemsChanged > 0 then
    if result then -- FIXME
        self:LayoutBags()
    end
end

function Sackville:GetVbagSectionForItem(item, bagSet, bagType)
-- item: The item we're trying to find a matching section for
-- bagSet: A table of virtual bags from which to choose matching section
-- bagType: which bag type we're allowed to match (player, bank, ...)
    self:Debug(3, "filter", "GetVbagSectionForItem: %s", item.name)
    
    local matchingSections = {}
    
    -- FIXME: iterate over the actual vbags
    for bagIndex=1, #self.Options.bagConfig do
        local bagOptions = self.Options.bagConfig[bagIndex]
        if bagType == bagOptions.bagType then
            local vbag = bagSet[bagOptions.name]
            if vbag then
                for sectionIndex=1, #bagOptions.sections do
                    local sectionName = bagOptions.sections[sectionIndex].name
                    assert (sectionName, sprintf("nil sectionName in bagOptions.sections[%d]", sectionIndex))
                    local section = vbag.sections[sectionName]
                    assert(section, sprintf("nil section (%s) in vbag (%s)", 
                        sectionName, vbag.name))
                    if Sackville.Filter:ItemMatchesSection(item, section) then
                        self:Debug(3, "filter", "> Item: %s, found matching section: %s with prio: %d",
                        item.name, section.name, section.priority)
                        table.insert(matchingSections, section)
                    end
                end
            else
                self:Debug(1, "filter", "!! nil vbag (%s), item: (%s)",
                    bagOptions.name, item.name)
                assert(vbag)

            end
        end
    end
    
    -- Sort by descending priority. Highest priority wins.
    for i, section in pairs(matchingSections) do
        self:Debug(3, "filter", "- Matching section: %s, prio: %d, index: %d",
            section.name, section.priority, i)
    end
    local matchComparer = function(a, b) return a.priority > b.priority end
    table.sort(matchingSections, matchComparer)
    
    return matchingSections[1]
end

function Sackville:BANKFRAME_OPENED()
    self:Debug(1, "event", "BANKFRAME_OPENED")
    isBankOpen = true
    
    if not bankInitalized then
        for _, bagid in ipairs(bagIndexes.bank) do
            local result = self:RebuildInventory(bagid)           
            for _, item in ipairs(result.itemsAdded) do
                local section = self:GetVbagSectionForItem(item, self.bankBags, "bank")
                if section then
                    section.vbag:AddItem(item, section.name)
                else
                    self:Debug(1, "filter", "Couldn't find a bag for %s", item.name)
                end
            end
        end
    end
    
    bankInitalized = true
    self:ShowBags()
end

function Sackville:BANKFRAME_CLOSED()
    self:Debug(1, "event", "BANKFRAME_CLOSED")
    
    if isBankOpen then
        for _, vbag in pairs(self.bankBags) do
            vbag:Hide()
        end    
    end    
    
    isBankOpen = false

end

function Sackville:PLAYERBANKSLOTS_CHANGED(event, arg1)
    self:Debug(1, "event", "PLAYERBANKSLOTS_CHANGED: %s", arg1)
    -- When (arg1 <= NUM_BANKGENERIC_SLOTS), arg1 is the index of the generic
    -- bank slot that changed. When (arg1 > NUM_BANKGENERIC_SLOTS), 
    -- (arg1 - NUM_BANKGENERIC_SLOTS) is the index of the equipped bank bag that changed. 
    local bagid
    if (arg1 <= NUM_BANKGENERIC_SLOTS) then -- generic bank slot changed
        self:BAG_UPDATE(event, -1)
    else
        -- bagid = arg1 - NUM_BANKGENERIC_SLOTS    
        -- self:BAG_UPDATE(event, bagid)
    end
end

function Sackville:BAG_UPDATE_COOLDOWN(event, cid)
    cid = cid or "?"
    self:Debug(1, "event", "BAG_UPDATE_COOLDOWN:", (cid or "?"))
    
    for i, item in pairs(self.items) do
        if item ~= nil then self:UpdateButtonCooldown(item, i) end
    end

end

function Sackville:UNIT_INVENTORY_CHANGED(e, unitid)
    if (loading) then return end
    self:Debug(1, "event", "UNIT_INVENTORY_CHANGED, unitid: %s", unitid)
end

function Sackville:DISPLAY_SIZE_CHANGED()
    self:Debug(1, "event", "DISPLAY_SIZE_CHANGED")
end

function Sackville:ITEM_LOCK_CHANGED(event, bagid, slot)
-- Desaturate item button when it is locked (probably because you're moving it)
    self:Debug(1, "event", "ITEM_LOCK_CHANGED")
    if not (bagid and slot) then return end
    local item = self.items[bagid .. '-' .. slot]
    if (item ~= nil) then
        local _, _, locked = GetContainerItemInfo(bagid, slot);
        SetItemButtonDesaturated(item.button, locked);
    end

end

function Sackville:UpdateMoney()
    self.quickPanel.money:SetText(Sackville.GetMoneyString())
end


function Sackville:ChatCommand(input)
    input = (input) and input or ""

    if not input or input:trim() == '' then
        self:ToggleBags()
    elseif input == 'bi' then
        self:RebuildInventory()
    elseif input == 'lay' then
        self:LayoutBags()
    end
end

function Sackville:ToggleBags()
    if showBags then
        self:HideBags()
    else
        self:ShowBags()
    end
end

function Sackville:ShowBags()
    self:Debug(1, "status", "ShowBags()")
    showBags = true
    Sackville:RegisterEvent("BAG_UPDATE")
	Sackville:RegisterEvent("ITEM_LOCK_CHANGED")
	Sackville:RegisterEvent("BAG_UPDATE_COOLDOWN")
	Sackville:RegisterEvent("DISPLAY_SIZE_CHANGED")
    
    self.quickPanel:Show()
    self:UpdateMoney()
    for _, vbag in pairs(self.playerBags) do
        vbag:Show()
    end
    
    if isBankOpen then
        for _, vbag in pairs(self.bankBags) do
            vbag:Show()
        end    
    end
    PlaySound("igBackPackOpen")

    self:LayoutBags()
end

function Sackville:HideBags()
    self:Debug(1, "status", "HideBags()")
    showBags = false
	Sackville:UnregisterEvent("BAG_UPDATE")
	Sackville:UnregisterEvent("ITEM_LOCK_CHANGED")
	Sackville:UnregisterEvent("BAG_UPDATE_COOLDOWN")
	Sackville:UnregisterEvent("DISPLAY_SIZE_CHANGED")
    
    self.quickPanel:Hide()
    for _, vbag in pairs(self.playerBags) do
        vbag:Hide()
    end
    
    if isBankOpen then
        for _, vbag in pairs(self.bankBags) do
            vbag:Hide()
        end    
    end    
    PlaySound("igBackPackClose")    
end

function Sackville:LayoutBags()
    self:Debug(1, "status", "LayoutBags()")

    local positionState = {
        x_offset = -20,
        y_offset = 80,
        widest_bag = 0
    }
    
    local bagSortOptions = {
        [1] = {
            key = "displayOrder",
            comparison = "<"
        }
    }
    
    positionState = PositionFrame(self.quickPanel, positionState)
    
    for _, vbag in self.Util.kpairs(self.playerBags, bagSortOptions) do
        vbag:LayoutItems()
        positionState = PositionFrame(vbag:GetFrame(), positionState)
    end
    
    if isBankOpen then
        for _, vbag in self.Util.kpairs(self.bankBags, bagSortOptions) do
            vbag:LayoutItems()
            positionState = PositionFrame(vbag:GetFrame(), positionState)
        end
    end
end

function Sackville:Debug(verbosity, flag, formatString, ...)
    if debugFlags[flag] then
        if debugVerbosity >= verbosity then
            self:Print(sprintf(formatString, ...))
        end
    end
end